;;
;; nnfido backend
;;
;; Copyright 2002 Pavel Kurnosoff, pavkur@hotbox.ru.
;;                                                  All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;; 3. All advertising materials mentioning features or use of this software
;;    must display the following acknowledgement:
;; 4. Neither the names of the authors nor the names of its contributors
;;    may be used to endorse or promote products derived from this software
;;    without specific prior written permission.
;;
;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;; SUCH DAMAGE.

(eval-when-compile (require 'cl))

(require 'nnoo)
(require 'message)
(require 'gnus-util)
(require 'gnus)
(require 'nnmail)

(defvar nnfido-arealist nil)
(defvar nnfido-process nil)
(defvar nnfido-group "")
(defvar nnfido-result "")
(defvar nnfido-eor "")

(defgroup gnus-nnfido nil
  "FidoNet backend for Gnus"
  :group 'gnus)

(defcustom nnfido-config "fidoconf///etc/fido/config" "Config file.\n\
The syntax is `type//path/to/config', where type is `squish', or `fidoconfig'
"
  :type 'string
  :group 'gnus-nnfido)

(defcustom nnfido-executable "/usr/local/bin/nnfido" "Path to nnfido executable"
  :type 'string
  :group 'gnus-nnfido)

(defcustom nnfido-default-aka "2:5030/1061" "Default AKA"
  :type 'string
  :group 'gnus-nnfido)

(defcustom nnfido-netmail-area "netmail" "Netmail area for posting"
  :type 'string
  :group 'gnus-nnfido)

(defun nnfido-sendln (command)
  (if nnfido-process
	  (process-send-string nnfido-process (concat command "\n"))))

(defun nnfido-fetch-result (&optional to-buffer)
  (set-process-buffer nnfido-process (or to-buffer nntp-server-buffer))
  (set-process-filter nnfido-process nil)
  (save-excursion
	(set-buffer (or to-buffer nntp-server-buffer))
	(erase-buffer)
	(while (and (not (= (or (char-before) ?a) ?\000))
				(memq (process-status nnfido-process) '(open run)))
	  (accept-process-output nnfido-process 1))
	(delete-char -1))
  (set-process-buffer nnfido-process nil)
  t)

(defun nnfido-get-result ()
	(setq nnfido-result "")
	(set-process-filter nnfido-process
						'(lambda (process output)
						   (setq nnfido-result (concat nnfido-result output))
						   ))
	(while (and (not (string-match "\000" nnfido-result)) (memq (process-status nnfido-process) '(open run)))
	  (accept-process-output nnfido-process 1))
	
	(substring nnfido-result 0 -1)
	)

(defun nnfido-cleanup ()
  (if (and nnfido-process (memq (process-status nnfido-process) '(open run)))
	  (progn (nnfido-sendln "EXIT")))
  (setq nnfido-process nil)
  (setq nnfido-arealist nil))

(nnoo-declare nnfido)
(nnoo-define-basics nnfido)

(deffoo nnfido-open-server (server &optional defenitions)
  (if (and nnfido-process (not (memq (process-status nnfido-process)
									 '(open run))))
	  (setq nnfido-process nil))	
  (unless nnfido-process
	(setq nnfido-process (let ((process-connection-type nil))
						   (start-process "nnfido" nil
										nnfido-executable))))
  (setq nnfido-arealist (nnfido-load-arealist nnfido-config))
  t)


(deffoo nnfido-retrieve-headers (articles &optional group server fetch-old)
  (nnfido-sendln (format "HEAD %s %s" (nnfido-getpath (or group nnfido-group)) articles))
  (nnfido-fetch-result)
  'nov)


(deffoo nnfido-request-group (group &optional server dont-check)
  (setq nnfido-group group)
  (unless dont-check
	(nnfido-sendln (format "STAT %s %s" (nnfido-getpath group) group))
	(nnfido-sendln "FINI")
	(let ((res (nnfido-get-result)) lowm hiwm)
	  (string-match "^[^ ]+ +\\([0-9]+\\) +\\([0-9]+\\)" res)
	  (setq hiwm (substring res (match-beginning 1) (match-end 1)))
	  (setq lowm (substring res (match-beginning 2) (match-end 2)))
	  (save-excursion
		(set-buffer nntp-server-buffer)
		(erase-buffer)
		(insert (format "211 %s %s %s %s" hiwm lowm hiwm group)))))
  t)

(deffoo nnfido-close-group (group &optional server)
  (setq nnfido-group nil))

(deffoo nnfido-request-article (article &optional group server buffer)
  (nnfido-sendln (format "RETR %s %d" (nnfido-getpath (or group nnfido-group)) article))
  (nnfido-fetch-result buffer)
  (save-excursion
	(set-buffer buffer)
	(goto-char (point-min))
	(insert (format "Newsgroups: %s\n" (or group nnfido-group))))
  t)


(deffoo nnfido-request-close ()
  (nnfido-cleanup))

(deffoo nnfido-close-server (&optional server)
  (nnfido-cleanup)
  (nnoo-close-server 'nnfido server))

(deffoo nnfido-retrieve-groups (groups &optional server)
  (set-process-buffer nnfido-process nntp-server-buffer)
  (set-process-filter nnfido-process nil)
  (save-excursion
	(set-buffer nntp-server-buffer)
	(erase-buffer)
	(dolist (area groups)
	  (setq area (assoc area nnfido-arealist))
	  (nnfido-sendln (format "STAT %s %s" (cadr area) (car area)))
	  (accept-process-output nnfido-process 0.01))
	(nnfido-sendln "FINI")
	(while (and (not (= (or (char-before) ?a) ?\000))
				(memq (process-status nnfido-process) '(open run)))
	  (accept-process-output nnfido-process 1))
	(delete-char -1))
  (set-process-buffer nnfido-process nil)
  'active)
  
(deffoo nnfido-request-list (&optional server)
  (set-process-buffer nnfido-process nntp-server-buffer)
  (set-process-filter nnfido-process nil)
  (save-excursion
	(set-buffer nntp-server-buffer)
	(erase-buffer)
	(dolist (area nnfido-arealist)
	  (nnfido-sendln (format "STAT %s %s" (cadr area) (car area)))
	  (accept-process-output nnfido-process 0.01))
	(nnfido-sendln "FINI")
	(while (and (not (= (or (char-before) ?a) ?\000))
				(memq (process-status nnfido-process) '(open run)))
	  (accept-process-output nnfido-process 1))
	(delete-char -1)
	)
  (set-process-buffer nnfido-process nil)
  t)

(deffoo nnfido-server-opened (&optional server)
  nnfido-process)

(deffoo nnfido-status-message (&optional server)
  "200 All Ok")

(defun nnfido-get-real-name (from)
  (if from
	  (if (string-match " *<.*> *" from)
		  (concat (substring from 0 (match-beginning 0))
				  (substring from (match-end 0)))
		(if (string-match "\(.*\)" from)
			(substring from (+ (match-beginning 0) 1) (- (match-end 0) 1))
		  from)) nil))

(defun nnfido-unhex-string (str)
  (with-string-as-buffer-contents str
	(goto-char 1)
	(while (re-search-forward "\\$\\([0-9A-F][0-9A-F]\\)" nil t)
	  (replace-match (char-to-string
					  (int-char (string-to-int (match-string 1) 16)))
					 t)
	  (buffer-string))))

(defun nnfido-get-reply (refs)
  (if refs
	  (progn (setq refs (car (last (split-string refs))))
			 (if (string-match "\\<\\(.*\\)@nnfido.exe\\>" refs)
				 (nnfido-unhex-string (substring refs (match-beginning 1) (match-end 1)))
			   (replace-in-string refs "\\<\\(.*\\)\\>" "\\1")))
	""))

(defun nnfido-get-ftn-address (to-field)
  (if to-field
	  (or
	   (and (string-match "@p[0-9]+\\.f[0-9]+\\.n[0-9]+\\.z[0-9]+\\." to-field)
			(replace-in-string to-field
							   "^.*@p\\([0-9]+\\)\\(\\.\\)f\\([0-9]+\\)\\.n\\([0-9]+\\)\\.z\\([0-9]+\\)\\..*$"
							   "\\5:\\4/\\3\\2\\1"
							   ))
	   (and (string-match "@f[0-9]+\\.n[0-9]+\\.z[0-9]+\\." to-field)
			(replace-in-string to-field
							   "^.*@f\\([0-9]+\\)\\.n\\([0-9]+\\)\\.z\\([0-9]+\\)\\..*$"
							   "\\3:\\2/\\1"
							   )))))   

(defun nnfido-parse-squish-config (path)
  (let (arealist areatag areapath extra aka)
	(save-excursion
	  (set-buffer (get-buffer-create "*squish-config*"))
	  (erase-buffer)
	  (insert-file-contents path)
	  (goto-char (point-min))
	  (while (re-search-forward "^[ \t]*\\(Echo\\|Net\\|Bad\\|Dupe\\)Area[ \t]+\\([^ \t]+\\)[ \t]+\\([^ \t]+\\)[ \t]+\\(.*\\)$" nil t)
		(setq areatag (buffer-substring (match-beginning 2) (match-end 2)))
		(setq areapath (buffer-substring (match-beginning 3) (match-end 3)))
		(setq extra (buffer-substring (match-beginning 4) (match-end 4)))
		(unless (string-match "-0" extra)
		  (setq aka nnfido-default-aka)
		  (if (string-match "-\\$" extra) (setq areapath (concat "$" areapath)))
		  (if (string-match "-p\\([^ \t]+\\)" extra)
			  (setq aka (substring extra (match-beginning 1) (match-end 1))))
		  (push (list areatag areapath aka) arealist))
		(forward-line 1))
	  (kill-buffer nil)
	  )
	arealist))

(defun nnfido-parse-fidoconf-config (path)
  (let (arealist areatag areapath extra aka)
	(save-excursion
	  (set-buffer (get-buffer-create "*fidoconf-config*"))
	  (erase-buffer)
	  (insert-file-contents path)
	  (goto-char (point-min))
	  (while (re-search-forward "^[ \t]*\\(Echo\\|Local\\|Netmail\\|Bad\\|Dupe\\)Area[ \t]+\\([^ \t]+\\)[ \t]+\\([^ \t]+\\)[ \t]+\\(.*\\)$" nil t)
		(setq areatag (buffer-substring (match-beginning 2) (match-end 2)))
		(setq areapath (buffer-substring (match-beginning 3) (match-end 3)))
		(setq extra (buffer-substring (match-beginning 4) (match-end 4)))
		(unless (string-match "-0" extra)
		  (setq aka nnfido-default-aka)
		  (if (string-match "-b[ \t]*Squish" extra) (setq areapath (concat "$" areapath)))
		  (if (string-match "-b[ \t]*Jam" extra) (setq areapath (concat "#" areapath)))
		  (if (string-match "-a\\([^ \t]+\\)" extra)
			  (setq aka (substring extra (match-beginning 1) (match-end 1))))
		  (push (list areatag areapath aka) arealist))
		(forward-line 1))
	  (kill-buffer nil)
	  )
	arealist))

(defun nnfido-parse-lisp-config (path)
  (save-excursion
	(set-buffer (get-buffer-create "*lisp-config*"))
	(erase-buffer)
	(insert-file-contents path)
	(prog1 (read (get-buffer "*lisp-config*"))
	  (kill-buffer nil))))


(defun nnfido-load-arealist (url)
  (if (string-match "^\\([a-zA-Z]+\\)//" url)
	  (let ((type (substring url 0 (match-end 1))) (path (substring url (match-end 0))))
		(cond
		 ((string= type "lisp") (nnfido-parse-lisp-config path))
		 ((string= type "squish") (nnfido-parse-squish-config path))
		 ((string= type "fidoconf") (nnfido-parse-fidoconf-config path))
		 (t nil)))))

(defun nnfido-getpath (area)
  (cadr (assoc area nnfido-arealist)))


(defun nnfido-message-goto-body ()
  (or (message-goto-body)
	  (progn 
		(goto-char (point-min))
		(search-forward "\n\n" nil t))))

(defun nnfido-do-post (area netmail)
  (let (body-begin ctl-info)
	(nnfido-message-goto-body)
	(setq body-begin (point))
	(setq ctl-info (concat "--- " (or (message-fetch-field "x-tearline")
									  (message-fetch-field "user-agent")
									  (message-fetch-field "x-newsreader")
									  (message-fetch-field "x-mailer")
									  "")
						   "\n * Origin: "
						   (or (message-fetch-field "organization") "")
						   " (" (caddr area) ")"
						   ))
	(nnfido-sendln
	 (format "POST %s %s %s %d"
			 (cadr area)
			 (caddr area)
			 (or
			  (nnfido-get-ftn-address (message-fetch-field "to"))
			  (nnfido-get-ftn-address (message-fetch-reply-field "from")))
			 (+ (- (point-max) body-begin) (length ctl-info) 1)))
	
	(nnfido-sendln (nnfido-get-reply (message-fetch-field "references")))
	
	(nnfido-sendln (or (nnfido-get-real-name (message-fetch-field "from"))
					   ""))
	(nnfido-sendln (replace-in-string
					(or
					 (message-fetch-field "x-comment-to")
					 (nnfido-get-real-name (message-fetch-field "to"))
					 (nnfido-get-real-name (message-fetch-reply-field "from"))
					 "All")
					"\"\\(.*\\)\"" "\\1"))
	(nnfido-sendln (replace-in-string (message-fetch-field "subject")
									  "Re\\(\\^[0-9]+\\)?: " ""))
	(process-send-region nnfido-process	body-begin (point-max))
	(nnfido-sendln ctl-info)
	))

(defun message-send-mail-with-nnfido ()
  (nnfido-do-post (assoc nnfido-netmail-area nnfido-arealist) t))

(deffoo nnfido-request-post (&optional server)
  (let (area)
	(setq area (assoc (car (split-string (message-fetch-field "newsgroups") ",[ \f\t\n\r\v]*"))
					  nnfido-arealist))
	(nnfido-do-post area nil)
	) t)

(provide 'nnfido)

(gnus-declare-backend "nnfido" 'address 'post)


